# http://interactivepython.org/runestone/static/pythonds/Recursion/recursioncomplex.html
# Use the turtle to find the exit path through a maze

import turtle

PART_OF_PATH = 'O'
TRIED = '.'
OBSTACLE = '+'
DEAD_END = '-'
EXIT_POINT = '*'

class Maze:
	def print_maze(self):
		s = ""
		num = 0
		for row in self.maze_list:
			s += "row: " + str(num) + ": "
			for col in row:
				s += str(col)
			s += "\n"
			num += 1
		print(s)
		
	def __init__(self, maze_filename):
		maze_file = open(maze_filename, 'r')
		row_number_in_maze = 0
		col_number_in_maze = 0
		self.maze_list = []

		for line in maze_file:
			rowList = []
			col = 0
			for ch in line[0: len(line)-1]:
				rowList.append(ch)
				if ch == 'S':
					self.start_row = row_number_in_maze
					self.start_col = col
				col = col + 1
			row_number_in_maze = row_number_in_maze + 1
			self.maze_list.append(rowList)
			col_number_in_maze = len(rowList)
			
		self.rows_in_maze = row_number_in_maze
		self.columns_in_maze = col_number_in_maze
		
		self.xTranslate = - col_number_in_maze / 2
		self.yTranslate = row_number_in_maze / 2
		self.print_maze()
		self.t = turtle.Turtle()
		self.t.shape('turtle')
		self.wn = turtle.Screen()

		#Sets the turtle world to be from -11, -5.5 to 11, 5.5
		self.wn.setworldcoordinates(-(col_number_in_maze-1)/2-.5, -(row_number_in_maze-1)/2-.5, (col_number_in_maze-1)/2+.5, (row_number_in_maze-1)/2+.5)
		
	def get_maze_size(self):
		return (self.rows_in_maze, self.columns_in_maze)
		
	def get_start_position(self):
		return (self.start_row, self.start_col)
		
	def draw_maze(self):
		self.t.speed(50) #was 10
		
		for y in range(self.rows_in_maze):
			for x in range(self.columns_in_maze):
				if self.maze_list[y][x] == OBSTACLE:
					self.draw_centered_box(x + self.xTranslate, -y + self.yTranslate, 'green')
					
		self.t.color('black')
		self.t.fillcolor('blue')

	def draw_centered_box(self, x, y,color):
		self.t.up()
		self.t.goto(x-.5, y-.5)
		self.t.color(color)
		self.t.fillcolor(color)
		self.t.setheading(90)
		self.t.down()
		
		self.t.begin_fill()
		for i in range(4):
			self.t.forward(1)
			self.t.right(90)
		self.t.end_fill()

	def move_turtle(self, x, y):
		self.t.up()
		self.t.setheading(self.t.towards(x + self.xTranslate, -y + self.yTranslate))
		self.t.goto(x + self.xTranslate, -y + self.yTranslate)

	def drop_breadcrumb(self, color):
		if color == "blue":
			#self.t.dot(40, color)
			self.t.color(color)
			self.t.stamp()
		else:
			self.t.dot(10, color)

	def  update_position(self, row, col, val=None):
		if val:
			self.maze_list[row][col] = val
			
		self.move_turtle(col, row)

		if val == PART_OF_PATH:
			print("part of path", row, col)
			color = 'yellow'
		elif val == TRIED:
			print("Tried", row, col)
			color = 'black'
		elif val == DEAD_END:
			print("dead end", row, col)
			color = 'red'
		elif val == EXIT_POINT:
			print("EXIT_POINT", row, col)
			color = 'blue'
		else:
			print("No val", row, col)
			color = None
		
		if color:
			self.drop_breadcrumb(color)

	def is_exit(self, row, col):
		result = (row == 0 or row == self.rows_in_maze-1 or col == 0 or col == self.columns_in_maze-1)
		return result

	def __getitem__(self, idx):
		return self.maze_list[idx]


def search_from(maze, current_row, current_column, rows_in_maze, cols_in_maze):
	# try each of four directions from this point until we find a way out.
	
	#  base case 1. row or column to try is outside the maze area
	if current_row >= rows_in_maze or current_row < 0 or current_column >= cols_in_maze or current_column < 0:
		return False
	
	#  base case 2. We have run into an obstacle, return false		
	if maze[current_row][current_column] == OBSTACLE:
		return False
		
    #  base case 3. We have found a square that has already been explored
	if maze[current_row][current_column] == TRIED or maze[current_row][current_column] == DEAD_END:
		return False
		
    #  base case 4. We have found an outside edge not occupied by an obstacle	
	if maze.is_exit(current_row, current_column):
		print("Hurray Hurray!! Have reached Exit!", current_row, current_column)
		maze. update_position(current_row, current_column, EXIT_POINT)
		maze.print_maze()
		return True
		
	maze. update_position(current_row, current_column, TRIED)
	
    # Otherwise, use logical short circuiting to try each of the four directions
    # in turn (if needed)
	try_north = search_from(maze, current_row-1, current_column, rows_in_maze, cols_in_maze)
	if try_north:
		return True
	try_south = search_from(maze, current_row+1, current_column, rows_in_maze, cols_in_maze)
	if try_south:
		return True
	try_east = search_from(maze, current_row, current_column+1, rows_in_maze, cols_in_maze) 
	if try_east:
		return True
	try_west = search_from(maze, current_row, current_column-1, rows_in_maze, cols_in_maze) 
	if try_west:
		return True

    # Tried all directions and none worked, therefore no exit from this point	
	maze. update_position(current_row, current_column, DEAD_END)
	return False

def main():
	my_maze = Maze('maze1.txt')
	my_maze.draw_maze()
	search_start = my_maze.get_start_position()
	maze_size = my_maze.get_maze_size()
	search_result = search_from(my_maze, search_start[0], search_start[1], maze_size[0], maze_size[1])
	
	turtle.mainloop()

main()
